@@ -34,8 +34,7 @@ import gitdb
 
 # typing ------------------------------------------------------
 
-from git.types import TBD, PathLike
-from typing_extensions import Literal
+from git.types import TBD, PathLike, Lit_config_levels
 from typing import (Any, BinaryIO, Callable, Dict,
                     Iterator, List, Mapping, Optional,
                     TextIO, Tuple, Type, Union,
@@ -46,7 +45,6 @@ if TYPE_CHECKING:  # only needed for types
     from git.refs.symbolic import SymbolicReference
     from git.objects import TagObject, Blob, Tree  # NOQA: F401
 
-Lit_config_levels = Literal['system', 'global', 'user', 'repository']
 
 # -----------------------------------------------------------
 
@@ -434,7 +432,7 @@ class Repo(object):
         """Delete the given tag references"""
         return TagReference.delete(self, *tags)
 
-    def create_remote(self, name: str, url: PathLike, **kwargs: Any) -> Remote:
+    def create_remote(self, name: str, url: str, **kwargs: Any) -> Remote:
         """Create a new remote.
 
         For more information, please see the documentation of the Remote.create
@@ -443,7 +441,7 @@ class Repo(object):
         :return: Remote reference"""
         return Remote.create(self, name, url, **kwargs)
 
-    def delete_remote(self, remote: 'Remote') -> Type['Remote']:
+    def delete_remote(self, remote: 'Remote') -> str:
         """Delete the given remote."""
         return Remote.remove(self, remote)
 
@@ -988,8 +986,6 @@ class Repo(object):
     def _clone(cls, git: 'Git', url: PathLike, path: PathLike, odb_default_type: Type[GitCmdObjectDB],
                progress: Optional[Callable], multi_options: Optional[List[str]] = None, **kwargs: Any
                ) -> 'Repo':
-        progress_checked = to_progress_instance(progress)
-
         odbt = kwargs.pop('odbt', odb_default_type)
 
         # when pathlib.Path or other classbased path is passed
@@ -1012,9 +1008,9 @@ class Repo(object):
         if multi_options:
             multi = ' '.join(multi_options).split(' ')
         proc = git.clone(multi, Git.polish_url(url), clone_path, with_extended_output=True, as_process=True,
-                         v=True, universal_newlines=True, **add_progress(kwargs, git, progress_checked))
-        if progress_checked:
-            handle_process_output(proc, None, progress_checked.new_message_handler(),
+                         v=True, universal_newlines=True, **add_progress(kwargs, git, progress))
+        if progress:
+            handle_process_output(proc, None, to_progress_instance(progress).new_message_handler(),
                                   finalize_process, decode_streams=False)
         else:
             (stdout, stderr) = proc.communicate()
